/*
 * Tiny Vector Matrix Library
 * Dense Vector Matrix Libary of Tiny size using Expression Templates
 *
 * Copyright (C) 2001 - 2003 Olaf Petzold <opetzold@users.sourceforge.net>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * $Id: VectorBinaryFunctions.h,v 1.7 2003/11/30 08:26:26 opetzold Exp $
 */

#ifndef TVMET_XPR_VECTOR_BINARY_FUNCTIONS_H
#define TVMET_XPR_VECTOR_BINARY_FUNCTIONS_H

namespace tvmet {


/*
 * binary_function(XprVector<E1, Sz>, XprVector<E2, Sz>)
 */
#define TVMET_IMPLEMENT_MACRO(FUNC, FUNCTIONAL)                 	\
template<class E1, class E2, std::size_t Sz>                   		\
inline                 			      			      	\
XprVector<                    	      					\
  XprBinOp<                    			      			\
    FUNCTIONAL<typename E1::value_type, typename E2::value_type>,	\
    XprVector<E1, Sz>,         			      			\
    XprVector<E2, Sz>          			      			\
  >,									\
  Sz									\
>                   			      		      		\
FUNC(const XprVector<E1, Sz>& lhs, const XprVector<E2, Sz>& rhs) {	\
  typedef XprBinOp<							\
    FUNCTIONAL<typename E1::value_type, typename E2::value_type>,	\
    XprVector<E1, Sz>,							\
    XprVector<E2, Sz>          			      			\
  >		    					expr_type; 	\
  return XprVector<expr_type, Sz>(		      			\
    expr_type(lhs, rhs));   						\
}

TVMET_IMPLEMENT_MACRO(atan2, Fcnl_atan2)
TVMET_IMPLEMENT_MACRO(drem, Fcnl_drem)
TVMET_IMPLEMENT_MACRO(fmod, Fcnl_fmod)
TVMET_IMPLEMENT_MACRO(hypot, Fcnl_hypot)
TVMET_IMPLEMENT_MACRO(jn, Fcnl_jn)
TVMET_IMPLEMENT_MACRO(yn, Fcnl_yn)
TVMET_IMPLEMENT_MACRO(pow, Fcnl_pow)

#undef TVMET_IMPLEMENT_MACRO


/*
 * binary_function(XprVector<E, Sz>, POD)
 */
#define TVMET_IMPLEMENT_MACRO(FUNC, FUNCTIONAL, TP)           		\
template<class E, std::size_t Sz>                  			\
inline                 			      			      	\
XprVector<                						\
  XprBinOp<          							\
    FUNCTIONAL<typename E::value_type, TP >,          			\
    XprVector<E, Sz>,         						\
    XprLiteral< TP >          						\
  >,               			      			      	\
  Sz               			      			      	\
>          								\
FUNC(const XprVector<E, Sz>& lhs, TP rhs) {				\
  typedef XprBinOp<          						\
    FUNCTIONAL<typename E::value_type, TP >, 				\
    XprVector<E, Sz>,  							\
    XprLiteral< TP >          						\
  >							expr_type;	\
  return XprVector<expr_type, Sz>(		      			\
    expr_type(lhs, XprLiteral< TP >(rhs))); 				\
}

TVMET_IMPLEMENT_MACRO(atan2, Fcnl_atan2, int)
TVMET_IMPLEMENT_MACRO(drem, Fcnl_drem, int)
TVMET_IMPLEMENT_MACRO(fmod, Fcnl_fmod, int)
TVMET_IMPLEMENT_MACRO(hypot, Fcnl_hypot, int)
TVMET_IMPLEMENT_MACRO(jn, Fcnl_jn, int)
TVMET_IMPLEMENT_MACRO(yn, Fcnl_yn, int)
TVMET_IMPLEMENT_MACRO(pow, Fcnl_pow, int)

#if defined(TVMET_HAVE_LONG_LONG)
TVMET_IMPLEMENT_MACRO(atan2, Fcnl_atan2, long long int)
TVMET_IMPLEMENT_MACRO(drem, Fcnl_drem, long long int)
TVMET_IMPLEMENT_MACRO(fmod, Fcnl_fmod, long long int)
TVMET_IMPLEMENT_MACRO(hypot, Fcnl_hypot, long long int)
TVMET_IMPLEMENT_MACRO(jn, Fcnl_jn, long long int)
TVMET_IMPLEMENT_MACRO(yn, Fcnl_yn, long long int)
TVMET_IMPLEMENT_MACRO(pow, Fcnl_pow, long long int)
#endif // defined(TVMET_HAVE_LONG_LONG)

TVMET_IMPLEMENT_MACRO(atan2, Fcnl_atan2, float)
TVMET_IMPLEMENT_MACRO(drem, Fcnl_drem, float)
TVMET_IMPLEMENT_MACRO(fmod, Fcnl_fmod, float)
TVMET_IMPLEMENT_MACRO(hypot, Fcnl_hypot, float)
TVMET_IMPLEMENT_MACRO(jn, Fcnl_jn, float)
TVMET_IMPLEMENT_MACRO(yn, Fcnl_yn, float)
TVMET_IMPLEMENT_MACRO(pow, Fcnl_pow, float)

TVMET_IMPLEMENT_MACRO(atan2, Fcnl_atan2, double)
TVMET_IMPLEMENT_MACRO(drem, Fcnl_drem, double)
TVMET_IMPLEMENT_MACRO(fmod, Fcnl_fmod, double)
TVMET_IMPLEMENT_MACRO(hypot, Fcnl_hypot, double)
TVMET_IMPLEMENT_MACRO(jn, Fcnl_jn, double)
TVMET_IMPLEMENT_MACRO(yn, Fcnl_yn, double)
TVMET_IMPLEMENT_MACRO(pow, Fcnl_pow, double)

#if defined(TVMET_HAVE_LONG_DOUBLE)
TVMET_IMPLEMENT_MACRO(atan2, Fcnl_atan2, long double)
TVMET_IMPLEMENT_MACRO(drem, Fcnl_drem, long double)
TVMET_IMPLEMENT_MACRO(fmod, Fcnl_fmod, long double)
TVMET_IMPLEMENT_MACRO(hypot, Fcnl_hypot, long double)
TVMET_IMPLEMENT_MACRO(jn, Fcnl_jn, long double)
TVMET_IMPLEMENT_MACRO(yn, Fcnl_yn, long double)
TVMET_IMPLEMENT_MACRO(pow, Fcnl_pow, long double)
#endif // defined(TVMET_HAVE_LONG_DOUBLE)

#undef TVMET_IMPLEMENT_MACRO


#if defined(TVMET_HAVE_COMPLEX)
/*
 * binary_function(XprVector<E, Sz>, std::complex<>)
 */
#define TVMET_IMPLEMENT_MACRO(FUNC, FUNCTIONAL)	           		\
template<class E, std::size_t Sz, class T>             			\
inline                 			      			      	\
XprVector<                						\
  XprBinOp<          							\
    FUNCTIONAL<typename E::value_type, std::complex<T> >, 		\
    XprVector<E, Sz>,         						\
    XprLiteral< std::complex<T> >   					\
  >,               			      			      	\
  Sz               			      			      	\
>          								\
FUNC(const XprVector<E, Sz>& lhs, const std::complex<T>& rhs) {		\
  typedef XprBinOp<          						\
    FUNCTIONAL<typename E::value_type, std::complex<T> >, 		\
    XprVector<E, Sz>,  							\
    XprLiteral< std::complex<T> >          				\
  >							expr_type;	\
  return XprVector<expr_type, Sz>(        				\
    expr_type(lhs, XprLiteral< std::complex<T> >(rhs))); 		\
}

TVMET_IMPLEMENT_MACRO(atan2, Fcnl_atan2)
TVMET_IMPLEMENT_MACRO(drem, Fcnl_drem)
TVMET_IMPLEMENT_MACRO(fmod, Fcnl_fmod)
TVMET_IMPLEMENT_MACRO(hypot, Fcnl_hypot)
TVMET_IMPLEMENT_MACRO(jn, Fcnl_jn)
TVMET_IMPLEMENT_MACRO(yn, Fcnl_yn)
TVMET_IMPLEMENT_MACRO(pow, Fcnl_pow)

#undef TVMET_IMPLEMENT_MACRO

#endif // defined(TVMET_HAVE_COMPLEX)


} // namespace tvmet

#endif // TVMET_XPR_VECTOR_BINARY_FUNCTIONS_H

// Local Variables:
// mode:C++
// End:
